
# bedrock-rnd - curated footprints, symbols and devmaps for Ringdove EDA
# Copyright (C) 2025  Tibor 'Igor2' Palinkas
#
# dist-license: GPL2+
# use-license (license of the output): unlimited (CC0 or public domain)
#
# For more information: doc/license.html
#        or on the web: http://www.repo.hu/projects/bedrock-rnd/license.html
#
# Project page: http://www.repo.hu/projects/bedrock-rnd
#

#@@doc-dwg centerpad-rfp.doc.png

#@@param:CPVP center pad via grid pitch (distance between vias); 0 or empty for no vias
#@@dim:CPVP

#@@param:CPVHD center pad via hole diameter
#@@dim:CPVHD

#@@param:CPVCD center pad via copper diameter
#@@dim:CPVCD

#@@param:CPPAT center pad pattern type
#@@enum:CPPAT:none         no pattern, round corner rectangle slightly smaller than the pad
#@@enum:CPPAT:chamfer      small segments with chamfer corners that avoid vias
#@@enum:CPPAT:nopaste      no paste at all
#@@default:CPPAT:chamfer

#@@param:CPPG center pad pattern gap; nominal distance between pattern elements
#@@dim:CPPG


BEGIN {
	centerpad_params="CPVP,CPVHD,CPVCD,CPPAT,CPPG"
}

function centerpad_setup(D, P)
{
	D["CPVP"] = parse_dim_(P["CPVP"], 0)
	D["CPVHD"] = either(parse_dim_(P["CPVHD"], 0), mm(0.3))
	D["CPVCD"] = either(parse_dim_(P["CPVCD"], 0), mm(0.8))
	D["CPPAT"] = P["CPPAT"]
	D["CPPG"] = either(parse_dim_(P["CPPG"], 0), mm(0.3))
}

function centerpad_has_pattern(D)
{
	return (D["CPPAT"] != "") && (D["CPPAT"] != "none")
}

function centerpad_has_vias(D)
{
	return (D["CPVP"] > 0) && (D["CPVHD"] > 0) && (D["CPVCD"] > 0)
}

# append a point to a polygo if it's not already appended
function cp_poly_append(POLY, len, xr, yr, x, y)
{
# clamp to pad
	if (x < -xr) x = -xr
	if (x > +xr) x = +xr
	if (y < -yr) y = -yr
	if (y > +yr) y = +yr

	if ((POLY[len-2] == x) && (POLY[len-1] == y)) {
		return
	}
	POLY[len++] = x
	POLY[len++] = y
	return len
}

function cp_poly_area(POLY, len     ,n,a)
{
	len -= 2
	for(n = 0; n < len; n+=2)
		a += POLY[n] * POLY[n+3] - POLY[n+1] * POLY[n+2]
	a += POLY[n] * POLY[1] - POLY[n+1] * POLY[0]
	if (a < 0) a = -a
	return a/2
}

function centerpad_pattern_chamfer(D, xmax, ymax, pit    ,x1,y1,x2,y2,xn,yn,xr,yr,gh,POLY,len,ch0,chv,xx1,xx2,yy1,yy2,min_area,paste_area,pa,comment)
{
	xr = D["cpad_width"]/2
	yr = D["cpad_length"]/2
	gh = D["CPPG"]/2

# chamfer at sharp corner
	ch0 = D["CPPG"]/8
	if (ch0 < 0.05) ch0 = 0.05

# chamfer at via hole
	chv = centerpad_has_vias(D) ? (D["CPVHD"] - D["CPPG"])/2+0.1 : ch0
	if (chv < 0.05) chv = 0.05

	min_area = (D["CPPG"]*D["CPPG"])

	for(yn = -ymax; yn <= ymax+1; yn++) {
		y1 = (yn-1) * pit + gh
		if (y1 < -yr) y1 = -yr;
		y2 = (yn) * pit - gh
		if (y2 > +yr) y2 = +yr;

		for(xn = -xmax; xn <= xmax+1; xn++) {
			x1 = (xn-1) * pit + gh
			if (x1 < -xr) x1 = -xr;
			x2 = (xn) * pit - gh
			if (x2 > +xr) x2 = +xr;

			if (((x2-x1) > D["CPPG"]/2) && ((y2-y1) > D["CPPG"]/2)) {
				len = 0;
				delete POLY

				# north side
				xx1 = x1 + (((x1 > -xr) && (y1 > -yr)) ? chv : ch0)
				xx2 = x2 - (((x2 < +xr) && (y1 > -yr)) ? chv : ch0)
				if (xx1 < xx2) {
					# normal
					len = cp_poly_append(POLY, len, xr, yr, xx1, y1)
					len = cp_poly_append(POLY, len, xr, yr, xx2, y1)
				}
				else {
					# flipped, would self intersect
					len = cp_poly_append(POLY, len, xr, yr, xx2, y1)
					len = cp_poly_append(POLY, len, xr, yr, xx1, y1)
				}

				# east side
				yy1 = y1 + (((x2 < +xr) && (y1 > -yr)) ? chv : ch0)
				yy2 = y2 - (((x2 < +xr) && (y2 < +yr)) ? chv : ch0)
				if (yy1 < yy2) {
					# normal
					len = cp_poly_append(POLY, len, xr, yr, x2, yy1)
					len = cp_poly_append(POLY, len, xr, yr, x2, yy2)
				}
				else {
					# flipped, would self intersect
					len = cp_poly_append(POLY, len, xr, yr, x2, yy2)
					len = cp_poly_append(POLY, len, xr, yr, x2, yy1)
				}

				# south side
				xx2 = x2 - (((x2 < +xr) && (y2 < +yr)) ? chv : ch0)
				xx1 = x1 + (((x1 > -xr) && (y2 < +yr)) ? chv : ch0)
				if (xx2 > xx1) {
					# normal
					len = cp_poly_append(POLY, len, xr, yr, xx2, y2)
					len = cp_poly_append(POLY, len, xr, yr, xx1, y2)
				}
				else {
					# flipped, would self intersect
					len = cp_poly_append(POLY, len, xr, yr, xx1, y2)
					len = cp_poly_append(POLY, len, xr, yr, xx2, y2)
				}

				# west side
				yy2 = y2 - (((x1 > -xr) && (y2 < +yr)) ? chv : ch0)
				yy1 = y1 + (((x1 > -xr) && (y1 > -yr)) ? chv : ch0)
				if (yy2 > yy1) {
					# normal
					len = cp_poly_append(POLY, len, xr, yr, x1, yy2)
					len = cp_poly_append(POLY, len, xr, yr, x1, yy1)
				}
				else {
					# flipped, would self intersect
					len = cp_poly_append(POLY, len, xr, yr, x1, yy1)
					len = cp_poly_append(POLY, len, xr, yr, x1, yy2)
				}

				area = cp_poly_area(POLY, len)
#print "area=", area, ">", min_area, ":", (area > min_area) > "/dev/stderr"
				if (area > min_area) {
					subc_poly("top-mask", POLY, 0)
					subc_poly("top-paste", POLY, 0)
					paste_area += area
				}
			}
		}
	}
	pa = D["cpad_width"] * D["cpad_length"]
#	print "paste_area=", paste_area " of " pa, paste_area/pa > "/dev/stderr"
	pa = paste_area/pa*100
	if (pa < 50)
		comment = "too little, use more paste!"
	else if (pa > 80)
		comment = "too much, use less paste!"
	else
		comment = "okay"
	print "Info: centerpad paste cover:", pa "%, which is", comment> "/dev/stderr"
}

function centerpad_pattern_nopaste(D, xmax, ymax, pit)
{
	subc_rect("top-mask", -D["cpad_width"]/2, -D["cpad_length"]/2, D["cpad_width"]/2, D["cpad_length"]/2)
}


function centerpad(D      ,want_pattern,want_vias,x,y,xn,yn,xmax,ymax,cpad_proto,via_proto,pit)
{
	want_pattern = centerpad_has_pattern(D);
	want_vias = centerpad_has_vias(D)

# centerpad copper
	if ((D["cpad_length"] > 0) && (D["cpad_width"] > 0)) {
		cid = D["cpad_termid"]
		if (cid == "")
			cid = "0"
		cpad_proto = ipc_subc_proto_begin()
		subc_pstk_add_shape_rsquare(cpad_proto, "top-copper", 0, 0, mm(D["cpad_width"]), mm(D["cpad_length"]), 0)
		if (!want_pattern) {
			subc_pstk_add_shape_rsquare(cpad_proto, "top-mask", 0, 0, mm(D["cpad_width"])+mask_offs*2, mm(D["cpad_length"])+mask_offs*2, 0)
			subc_pstk_add_shape_rsquare(cpad_proto, "top-paste", 0, 0, mm(D["cpad_width"])-0.1, mm(D["cpad_length"])-0.1, 0.1)
		}
#		else mask and paste are drawn by the patten code

		ipc_subc_proto_end(cpad_proto)
		subc_pstk(cpad_proto, 0, 0, 0, cid, cid, 0.2)
	}

	if (want_vias || want_pattern) {
		pit = D["CPVP"]
		if (pit == "")
			pit = mm(1.2)
		xmax = int(D["cpad_width"] / pit - 0.5)/2
		ymax = int(D["cpad_length"] / pit - 0.5)/2
	}

	if (want_vias) {
		via_proto = ipc_subc_proto_begin()
		subc_pstk_add_shape_circ(via_proto, "bottom-copper", 0, 0, D["CPVCD"])
		subc_pstk_add_shape_circ(via_proto, "intern-copper", 0, 0, D["CPVCD"])
		ipc_subc_proto_end(via_proto,  D["CPVHD"], 1)

		for(yn = -ymax; yn <= ymax; yn++) {
			y = yn * pit
			for(xn = -xmax; xn <= xmax; xn++) {
				x = xn * pit
				subc_pstk(via_proto, x, y, 0, cid, cid, 0.2)
			}
		}
	}

	if (want_pattern) {
		# xmax and ymax are set above when want_vias
		if (D["CPPAT"] == "chamfer")
			centerpad_pattern_chamfer(D, xmax, ymax, pit)
		else if (D["CPPAT"] == "nopaste")
			centerpad_pattern_nopaste(D, xmax, ymax, pit)
		else {
			print "Invalid (unknown) center pad pattern (CPPAT): ", D["CPPAT"] > "/dev/stderr"
			exit 1
		}
	}
}

