
# bedrock-rnd - curated footprints, symbols and devmaps for Ringdove EDA
# Copyright (C) 2025  Tibor 'Igor2' Palinkas
#
# dist-license: GPL2+
# use-license (license of the output): unlimited (CC0 or public domain)
#
# For more information: doc/license.html
#        or on the web: http://www.repo.hu/projects/bedrock-rnd/license.html
#
# Project page: http://www.repo.hu/projects/bedrock-rnd
#

#@@param:F fab tolerance in +-mm
#@@default:F 0.1

#@@param:P part placement tolerance in +-mm
#@@default:P 0.05

#@@param:silk_thickness thickness of silk graphic lines; diametr of pin 1 dot is calculated from this too

#@@param:mask_offs gap between copper pad edge and mask cutout edge

#@@param:glue_thick default line thickness for glue on adhesive layers; if 0 omit the layer

#@@param:glue_dots print dots instead of lines for glue when non-zero

#@@param:pad_rounding rectangular pad corner rounding radius (0 for sharp corners)
#@@dim:pad_rounding
#@@optional:pad_rounding

BEGIN {
	ipc_params=",silk_thickness,mask_offs,glue_thick,glue_dots"
}

# increase radius by offs and return, but if radius is 0 preserve it as 0 for
# sharp corner
function inc_radius(r, offs)
{
	if (r+0 <= 0)
		return 0
	return r+offs
}

# update the tolerance value for a parameter
function mktol(name)
{
	TOL[name] = MAX[name] - MIN[name]
	if (MAX[name] != MIN[name])
		NOM[name] = (MAX[name] + MIN[name]) / 2
	else
		NOM[name] = MAX[name]
}

# round val to the closest N*prec value
function round_to(val, prec)
{
	return int(val/prec+0.500001)*prec
}

function load_rfp(name, val)
{
	MIN[name] = val
	NOM[name] = val
	MAX[name] = val
	TOL[name] = 0
}

function copy_dim(dst, src)
{
	MIN[dst] = MIN[src]
	NOM[dst] = NOM[src]
	MAX[dst] = MAX[src]
	TOL[dst] = TOL[src]
}


# convert user specified dimension+tolerance into MIN[]/MAX[]/TOL[]
function load_dim(str, name    ,A,t,mul)
{
	if (str == "")
		return

	if (str ~ "inch$") {
		mul = 1000 / 39.3701
		sub("inch$", "", str)
	}
	else if (str ~ "mil$") {
		mul = 1 / 39.3701
		sub("mil$", "", str)
	}
	else if (str ~ "mm$") {
		mul = 1
		sub("mm$", "", str)
	}
	else
		mul = 1

	if (str ~ "~") {
		split(str, A, "[~]")
		A[1] = A[1]+0
		A[2] = A[2]+0
		if (mul != 1) {
			A[1] = A[1] * mul
			A[2] = A[2] * mul
		}
		MIN[name] = A[1]
		MAX[name] = A[2]
	}
	else if (str ~ ".[+].") {
		split(str, A, "[+]")
		A[1] = A[1]+0
		A[2] = A[2]+0
		if (mul != 1) {
			A[1] = A[1] * mul
			A[2] = A[2] * mul
		}
		if (A[2] < 0) {
			# +- form
			t = 0-A[2]
			MIN[name] = A[1] - t
			MAX[name] = A[1] + t
		}
		else {
			t = A[2]
			MIN[name] = A[1]
			MAX[name] = A[1] + t
		}
	}
	else {
		# exact value
		t = str*mul
		MIN[name] = t
		MAX[name] = t
	}
	mktol(name)
}


# initialize generic tolerances: F for fab and P for placement tolerenace
function ipc_init_gentol() {
	gentol_F = P["F"]
	gentol_P = P["P"]
	silk_thick = P["silk_thickness"]
	mask_offs = P["mask_offs"]
	glue_thick = P["glue_thick"]
	glue_dots = P["glue_dots"]
	pad_rounding = either(parse_dim(P["pad_rounding"]), mm(0.1))

# assumed fabrication tolerance
	if (gentol_F == "")
		gentol_F = 0.1

# assumed placement tolerance
	if (gentol_P == "")
		gentol_P = 0.05

	if (silk_thick == "")
		silk_thick = 0.2

# IPC says anything between 0 and 0.1 works; jlcpcb wants >=0.038
	if (mask_offs == "") {
		if (P["density"] == "C")
			mask_offs = 0.045
		else
			mask_offs = 0.075
	}
}



# Return rounded maximum pad extent (outer dimension); extent_name is package's
# extent MIN/MAX/TOL
function ipc_calc_extent(PADGEO, extent_name,            Cl,Zmax,jt)
{
# 3.1.1.1 and 3.1.5: Z is max pin extent (PE)
	if (PADGEO["reverse"])
		jt = PADGEO["JH"]
	else
		jt = PADGEO["JT"]

	Cl = TOL[extent_name]
	Zmax = MIN[extent_name] + 2*jt + sqrt(Cl*Cl + gentol_F*gentol_F + gentol_P*gentol_P)
	return round_to(Zmax, PADGEO["JTr"])
}


# Return rounded minimum gap extent (inner dimension); extent_name is package's
# extent MIN/MAX/TOL, gap_name is package's gap MIN/MAX/TOL
function ipc_calc_gap(PADGEO, extent_name, gap_name,     Stol_rms,Cl,Gmin,jh)
{
# 3.1.1: S is pin gap (PG)
	Stol_rms = sqrt(TOL[extent_name]*TOL[extent_name] + 2 * TOL["PC"]*TOL["PC"])
#	Smin = MIN[gap_name]
#	Smax_rms = Smin + Stol_rms
#print "Smax=" Smax " = ", Smin, "rms=" Stol_rms, " TPE=" TOL[extent_name], "TPC=" TOL["PC"] " -> ", Stol_rms > "/dev/stderr"

# 3.1.5
	if (PADGEO["reverse"])
		jt = PADGEO["JT"]
	else
		jt = PADGEO["JH"]

	Cs = Stol_rms
	Gmin = MAX[gap_name] - 2*jt - sqrt(Cs*Cs + gentol_F*gentol_F + gentol_P*gentol_P)
#print "Gmin=", Gmin, " = " MAX[gap_name], "-" 2*jt, "-" sqrt(Cs*Cs + gentol_F*gentol_F + gentol_P*gentol_P) > "/dev/stderr"

	return round_to(Gmin, PADGEO["JHr"]);
}


# Return rounded pad thickness; pin_thick_name is package's pin/lead
# thickness MIN/MAX/TOL
function ipc_calc_pin_thick(PADGEO, pin_thick_name     ,Cw,Xmax)
{
	Cw = TOL[pin_thick_name]
	Xmax = MIN[pin_thick_name] + 2*PADGEO["JS"] + sqrt(Cw*Cw + gentol_F*gentol_F + gentol_P*gentol_P)
#print "## Xmax=" Xmax, "round=" round_to(Xmax + PADGEO["JSr"], PADGEO["JSr"]*2) > "/dev/stderr"

# round to *2 because there are two sides but we return a single value
# including both roundings
	return round_to(Xmax+PADGEO["JSr"], PADGEO["JSr"]*2);
}

### drawing ###

function silk_line(x1, y1, x2, y2, thick)
{
	if (thick == "")
		thick = silk_thick
	subc_line("top-silk", x1, y1, x2, y2, thick)
}

function silk_pin1(x, y, thick)
{
	if (thick == "")
		thick = silk_thick
	subc_line("top-silk", x, y, x, y, thick*1.5)
}

function adhesive_line(x1, y1, x2, y2)
{
	if (glue_thick > 0) {
		if (glue_dots) {
			subc_line("top-adhesive", x1, y1, x1, y1, glue_thick)
			subc_line("top-adhesive", x2, y2, x2, y2, glue_thick)
		}
		else
			subc_line("top-adhesive", x1, y1, x2, y2, glue_thick)
	}
}

function asm_line(x1, y1, x2, y2, thick)
{
	if (thick == "")
		thick = 0.05
	subc_line("top-assy", x1, y1, x2, y2, thick)
}

function asm_box(x1, y1, x2, y2)
{
	asm_line(x1, y1, x2, y1)
	asm_line(x2, y1, x2, y2)
	asm_line(x2, y2, x1, y2)
	asm_line(x1, y2, x1, y1)
}

function asm_body_box(x1, y1, x2, y2  ,offs,offs_north)
{
	if (offs == "")
		offs = 0.1

	if (offs_north == "")
		offs_north = offs

	asm_box(x1, y1, x2, y2)
	asm_box(x1+offs, y1+offs_north, x2-offs, y2-offs)

	asm_line(x1, y1, x1+offs, y1+offs_north)
	asm_line(x2, y1, x2-offs, y1+offs_north)
	asm_line(x1, y2, x1+offs, y2-offs)
	asm_line(x2, y2, x2-offs, y2-offs)
}


### courtyard ###
BEGIN {
	min_api = 20241231
	pcb_rnd_minver= "3.1.7"
	if (common_subc_api_version < min_api) {
		if (common_subc_api_version == "")
			common_subc_api_version = "pre-2024";
		print "\ncommon_subc.awk is too old: required " min_api " but got " common_subc_api_version "\nPlease upgrade your pcb-rnd to at least " pcb_rnd_minver "\n" > "/dev/stderr"
		exit 1
	}
	base_unit_mm = 1
	cyrect = 0
	cy_rxy = ""
}

#function cy_line(x1, y1, x2, y2)
#{
#	print "line secondary silk -", x1, y1, x2, y2, "0.02 0.0000"
#}

function courtyard_add_rect(x1, y1, x2, y2)
{
	cy_rxy = cy_rxy " -rxy " q x1 "," y1 "," x2 "," y2 q
}

function courtyard_print(     cmd,s,l)
{
	if (cy_rxy == "")
		return

	cmd = "rect-union-rnd" cy_rxy

	s = ""
	close(cmd)
	while((cmd | getline l) == 1)
		s = s l NL
	close(cmd)

	if (s == "") {
		print "ERROR: failed to run rect-union-rnd; it should have been installed on $PATH when pcb-rnd got installed" > "/dev/stderr"
		exit 1
	}

	LAYER["top-courtyard"] = LAYER["top-courtyard"] NL s
}


### helpers ###

function ipc_subc_layers()
{
	layer_setup("top-paste", "top", "paste", 0, 1)
	layer_setup("top-mask",  "top", "mask", 1, 1)
	layer_setup("top-assy",  "top", "doc", 0, 0); LAYER_PURPOSE["top-assy"] = "assy"
	layer_setup("top-courtyard",  "top", "doc", 0, 0); LAYER_PURPOSE["top-courtyard"] = "ko.courtyard"
	if (glue_thick > 0)
		layer_setup("top-adhesive",  "top", "mech", 0, 0); LAYER_PURPOSE["top-adhesive"] = "adhesive"
}

function ipc_subc_begin(footprint, refdes, refdes_x, refdes_y, refdes_dir, SUBC_ATTR)
{
	if ((generator == "") || (gversion == "") || (bedrock_ver == "")) {
		print "ERROR: ipc_subc_begin(): need all three of these variables: generator, gversion, bedroc_ver" > "/dev/stderr"
		exit 1
	}

	SUBC_ATTR["generator"] = generator "::" gversion
	SUBC_ATTR["standard"] = "bedrock-rnd::" bedrock_ver
	SUBC_ATTR["license/use"] = "unlimited"
	SUBC_ATTR["license/dist"] = "unlimited"
	SUBC_ATTR["density"] = P["density"]
	if (P["datasheet"] != "")
		SUBC_ATTR["datasheet"] = P["datasheet"]

	if (footprint == "")
		footprint = generator "(" args ")"

	return subc_begin(footprint, refdes, refdes_x, refdes_y, refdes_dir, SUBC_ATTR)
}

function ipc_subc_proto_begin(     tmp)
{
	tmp = subc_proto_alloc()
	PROTO[tmp] = PROTO[tmp] "     li:shape {" NL
	return tmp
}

function ipc_subc_proto_end(pad_proto,  hdia, hplated)
{
	PROTO[pad_proto] = PROTO[pad_proto] "     }" NL

	if (hdia != "")
		subc_pstk_add_hole(pad_proto, hdia, hplated)
	else
		subc_pstk_no_hole(pad_proto)
}


### tables ###

# Fill in PADGEO[] with the solder join fillet tuning parameters
function pad_geo_gull_wing(PADGEO, density, pitch)
{
	# table 3-2 and 3-3 on page 22 of IPC-7351
	# TODO: this ignores footnote; Amax is in qfn and means maximum side-to-side
	#       extent of the perpendicular row of pads; see:
	#       https://web.archive.org/web/20241229045420/https://www.ti.com/lit/an/slua271c/slua271c.pdf?ts=1735373683137
	if (pitch > 0.625) {
		PADGEO["JTr"] = 0.2
		PADGEO["JHr"] = 0.2
		PADGEO["JSr"] = 0.05
		if (density == "C") {
			PADGEO["JT"] = 0.15
			PADGEO["JH"] = 0.25
			PADGEO["JS"] = 0.01
			PADGEO["cy"] = 0.1
		}
		else if (density == "B") {
			PADGEO["JT"] = 0.35
			PADGEO["JH"] = 0.35
			PADGEO["JS"] = 0.03
			PADGEO["cy"] = 0.25
		}
		else if (density == "A") {
			PADGEO["JT"] = 0.55
			PADGEO["JH"] = 0.45
			PADGEO["JS"] = 0.05
			PADGEO["cy"] = 0.5
		}
	}
	else {
		PADGEO["JTr"] = 0.2
		PADGEO["JHr"] = 0.1
		PADGEO["JSr"] = 0.05
		if (density == "C") {
			PADGEO["JT"] = 0.15
			PADGEO["JH"] = 0.25
			PADGEO["JS"] = -0.04
			PADGEO["cy"] = 0.1
		}
		else if (density == "B") {
			PADGEO["JT"] = 0.35
			PADGEO["JH"] = 0.35
			PADGEO["JS"] = -0.02
			PADGEO["cy"] = 0.25
		}
		else if (density == "A") {
			PADGEO["JT"] = 0.55
			PADGEO["JH"] = 0.45
			PADGEO["JS"] = 0.01
			PADGEO["cy"] = 0.5
		}
	}
}

function pad_geo_flat_ribbon_l(PADGEO, density, pitch)
{
	return pad_geo_gull_wing(PADGEO, density, pitch)
}

function pad_geo_J_leads(PADGEO, density)
{
	# table 3-4 on page 23 of IPC-7351
	PADGEO["reverse"] = 1
	PADGEO["JTr"] = 0.2
	PADGEO["JHr"] = 0.2
	PADGEO["JSr"] = 0.05
	if (density == "C") {
		PADGEO["JH"] = 0.15
		PADGEO["JT"] = -0.30
		PADGEO["JS"] = 0.01
		PADGEO["cy"] = 0.1
	}
	else if (density == "B") {
		PADGEO["JH"] = 0.35
		PADGEO["JT"] = -0.20
		PADGEO["JS"] = 0.03
		PADGEO["cy"] = 0.25
	}
	else if (density == "A") {
		PADGEO["JH"] = 0.55
		PADGEO["JT"] = -0.10
		PADGEO["JS"] = 0.05
		PADGEO["cy"] = 0.5
	}
}

# caps&resistors >= 0603 (1608 metric)
function pad_geo_rect_square_large(PADGEO, density)
{
	# table 3-5 on page 23 of IPC-7351
	PADGEO["JTr"] = 0.2
	PADGEO["JHr"] = 0.2
	PADGEO["JSr"] = 0.05
	if (density == "C") {
		PADGEO["JT"] = 0.15
		PADGEO["JH"] = -0.05
		PADGEO["JS"] = -0.05
		PADGEO["cy"] = 0.1
	}
	else if (density == "B") {
		PADGEO["JT"] = 0.35
		PADGEO["JH"] = -0.05
		PADGEO["JS"] = 0.00
		PADGEO["cy"] = 0.25
	}
	else if (density == "A") {
		PADGEO["JT"] = 0.55
		PADGEO["JH"] = -0.05
		PADGEO["JS"] = 0.05
		PADGEO["cy"] = 0.5
	}
}

# caps&resistors < 0603 (1608 metric)
function pad_geo_rect_square_small(PADGEO, density)
{
	# table 3-6 on page 23 of IPC-7351
	PADGEO["JTr"] = 0.05
	PADGEO["JHr"] = 0.05
	PADGEO["JSr"] = 0.05
	if (density == "C") {
		PADGEO["JT"] = 0.00
		PADGEO["JH"] = -0.05
		PADGEO["JS"] = 0.00
		PADGEO["cy"] = 0.1
	}
	else if (density == "B") {
		PADGEO["JT"] = 0.1
		PADGEO["JH"] = -0.05
		PADGEO["JS"] = 0.00
		PADGEO["cy"] = 0.15
	}
	else if (density == "A") {
		PADGEO["JT"] = 0.20
		PADGEO["JH"] = -0.05
		PADGEO["JS"] = 0.05
		PADGEO["cy"] = 0.20
	}
}

function pad_geo_rect_square(PADGEO, density, PE)
{
	if (PE > 1.6) {
		PADGEO["paste_reduction"] = 1
		return pad_geo_rect_square_large(PADGEO, density)
	}
	else {
		PADGEO["paste_reduction"] = 0
		return pad_geo_rect_square_small(PADGEO, density)
	}
}


# Cylindicral End Cap Terminations (MELF)
function pad_geo_cylindicral(PADGEO, density)
{
	# table 3-7 on page 23 of IPC-7351
	PADGEO["JTr"] = 0.2
	PADGEO["JHr"] = 0.2
	PADGEO["JSr"] = 0.1
	if (density == "C") {
		PADGEO["JT"] = 0.2
		PADGEO["JH"] = 0.02
		PADGEO["JS"] = 0.01
		PADGEO["cy"] = 0.1
	}
	else if (density == "B") {
		PADGEO["JT"] = 0.4
		PADGEO["JH"] = 0.1
		PADGEO["JS"] = 0.05
		PADGEO["cy"] = 0.25
	}
	else if (density == "A") {
		PADGEO["JT"] = 0.6
		PADGEO["JH"] = 0.2
		PADGEO["JS"] = 0.1
		PADGEO["cy"] = 0.5
	}
}

# Bottom Only Terminations
function pad_geo_bottom_only(PADGEO, density)
{
	# table 3-8 on page 23 of IPC-7351
	PADGEO["JTr"] = 0.2
	PADGEO["JHr"] = 0.1
	PADGEO["JSr"] = 0.05
	if (density == "C") {
		PADGEO["JT"] = 0.15
		PADGEO["JH"] = 0.25
		PADGEO["JS"] = 0.01
		PADGEO["cy"] = 0.1
	}
	else if (density == "B") {
		PADGEO["JT"] = 0.35
		PADGEO["JH"] = 0.35
		PADGEO["JS"] = 0.03
		PADGEO["cy"] = 0.25
	}
	else if (density == "A") {
		PADGEO["JT"] = 0.55
		PADGEO["JH"] = 0.45
		PADGEO["JS"] = 0.08
		PADGEO["cy"] = 0.5
	}
}

# Leadless Chip Carrier with Castellated Terminations
function pad_geo_leadless(PADGEO, density)
{
	# table 3-9 on page 23 of IPC-7351
	PADGEO["reverse"] = 1
	PADGEO["JTr"] = 0.2
	PADGEO["JHr"] = 0.2
	PADGEO["JSr"] = 0.05
	if (density == "C") {
		PADGEO["JH"] = 0.45
		PADGEO["JT"] = 0.05
		PADGEO["JS"] = -0.15
		PADGEO["cy"] = 0.1
	}
	else if (density == "B") {
		PADGEO["JH"] = 0.55
		PADGEO["JT"] = 0.15
		PADGEO["JS"] = 0.05
		PADGEO["cy"] = 0.25
	}
	else if (density == "A") {
		PADGEO["JH"] = 0.65
		PADGEO["JT"] = 0.25
		PADGEO["JS"] = 0.05
		PADGEO["cy"] = 0.5
	}
}

# Butt Joints
function pad_geo_butt(PADGEO, density)
{
	# table 3-10 on page 24 of IPC-7351
	PADGEO["JTr"] = 0.2
	PADGEO["JHr"] = 0.1
	PADGEO["JSr"] = 0.05
	if (density == "C") {
		PADGEO["JT"] = 0.6
		PADGEO["JH"] = 0.6
		PADGEO["JS"] = 0.1
		PADGEO["cy"] = 0.2
	}
	else if (density == "B") {
		PADGEO["JT"] = 0.8
		PADGEO["JH"] = 0.8
		PADGEO["JS"] = 0.2
		PADGEO["cy"] = 0.8
	}
	else if (density == "A") {
		PADGEO["JT"] = 1.0
		PADGEO["JH"] = 1.0
		PADGEO["JS"] = 0.3
		PADGEO["cy"] = 1.5
	}
}

# Inward Flat Ribbon L and Gull-Wing Leads (Tantalum Capacitors)
function pad_geo_inward_flat(PADGEO, density)
{
	# table 3-11 on page 24 of IPC-7351
	PADGEO["JTr"] = 0.2
	PADGEO["JHr"] = 0.2
	PADGEO["JSr"] = 0.1
	if (density == "C") {
		PADGEO["JT"] = 0.07
		PADGEO["JH"] = 0.2
		PADGEO["JS"] = -0.1
		PADGEO["cy"] = 0.1
	}
	else if (density == "B") {
		PADGEO["JT"] = 0.15
		PADGEO["JH"] = 0.5
		PADGEO["JS"] = -0.05
		PADGEO["cy"] = 0.25
	}
	else if (density == "A") {
		PADGEO["JT"] = 0.25
		PADGEO["JH"] = 0.8
		PADGEO["JS"] = 0.01
		PADGEO["cy"] = 0.5
	}
}

# Flat Lug Leads
function pad_geo_flat_lug(PADGEO, density)
{
	# table 3-12 on page 24 of IPC-7351
	PADGEO["JTr"] = 0.2
	PADGEO["JHr"] = 0.1
	PADGEO["JSr"] = 0.05
	if (density == "C") {
		PADGEO["JT"] = 0.15
		PADGEO["JH"] = 0.25
		PADGEO["JS"] = -0.04
		PADGEO["cy"] = 0.1
	}
	else if (density == "B") {
		PADGEO["JT"] = 0.35
		PADGEO["JH"] = 0.35
		PADGEO["JS"] = 0.01
		PADGEO["cy"] = 0.25
	}
	else if (density == "A") {
		PADGEO["JT"] = 0.55
		PADGEO["JH"] = 0.45
		PADGEO["JS"] = 0.05
		PADGEO["cy"] = 0.5
	}
}

# Flat, No Lead
function pad_geo_flat_no_lead(PADGEO, density)
{
	# table 3-13 on page 24 of IPC-7351
	# TODO: Note is not implemented yet (JH trimming)
	PADGEO["JTr"] = 0.05
	PADGEO["JHr"] = 0.05
	PADGEO["JSr"] = 0.05
	if (density == "C") {
		PADGEO["JT"] = 0.2
		PADGEO["JH"] = 0.0
		PADGEO["JS"] = -0.04
		PADGEO["cy"] = 0.1
	}
	else if (density == "B") {
		PADGEO["JT"] = 0.3
		PADGEO["JH"] = 0.0
		PADGEO["JS"] = -0.04
		PADGEO["cy"] = 0.25
	}
	else if (density == "A") {
		PADGEO["JT"] = 0.4
		PADGEO["JH"] = 0.0
		PADGEO["JS"] = -0.04
		PADGEO["cy"] = 0.5
	}
}

# Small Outline (SO), No-Lead
function pad_geo_so_no_lead(PADGEO, density)
{
	# table 3-14 on page 24 of IPC-7351
	PADGEO["JTr"] = 0.05
	PADGEO["JHr"] = 0.05
	PADGEO["JSr"] = 0.05
	if (density == "C") {
		PADGEO["JT"] = 0.20
		PADGEO["JH"] = 0.0
		PADGEO["JS"] = -0.04
		PADGEO["cy"] = 0.1
	}
	else if (density == "B") {
		PADGEO["JT"] = 0.3
		PADGEO["JH"] = 0.0
		PADGEO["JS"] = -0.04
		PADGEO["cy"] = 0.25
	}
	else if (density == "A") {
		PADGEO["JT"] = 0.4
		PADGEO["JH"] = 0.0
		PADGEO["JS"] = -0.04
		PADGEO["cy"] = 0.5
	}
}
